@@ -1,10 +1,12 @@
 /* -*- mode: C; c-basic-offset: 4; indent-tabs-mode: nil; -*- */
+#include <string.h>
+
 #include "hippo-canvas-internal.h"
 #include "hippo-canvas-item.h"
 #include "hippo-canvas-container.h"
 #include "hippo-canvas-marshal.h"
 
-static void     hippo_canvas_item_base_init (void                  *klass);
+static void     hippo_canvas_item_class_init (void *g_iface);
 
 enum {
     DESTROY,
@@ -14,6 +16,7 @@ enum {
     BUTTON_PRESS_EVENT,
     BUTTON_RELEASE_EVENT,
     MOTION_NOTIFY_EVENT,
+    SCROLL_EVENT,
     KEY_PRESS_EVENT,
     ACTIVATED,
     TOOLTIP_CHANGED,
@@ -26,178 +29,209 @@ hippo_canvas_item_get_type(void)
 {
     static GType type = 0;
     if (type == 0) {
-        static const GTypeInfo info =
-            {
-                sizeof(HippoCanvasItemIface),
-                hippo_canvas_item_base_init,
-                NULL /* base_finalize */
-            };
-        type = g_type_register_static(G_TYPE_INTERFACE, "HippoCanvasItem",
-                                      &info, 0);
+        type = g_type_register_static_simple (G_TYPE_INTERFACE,
+                                              _("HippoCanvasItem"),
+                                              sizeof (HippoCanvasItemIface),
+                                              (GClassInitFunc) hippo_canvas_item_class_init,
+                                              0, NULL, 0);
     }
 
     return type;
 }
 
 static void
-hippo_canvas_item_base_init(void *klass)
+hippo_canvas_item_class_init(void *g_iface)
 {
-    static gboolean initialized = FALSE;
-
-   if (!initialized) {
-        /* create signals in here */
-
-        /**
-         * HippoCanvasItem::destroy
-         *
-         * This canvas item has explicitely been destroyed. All references it holds
-         * to other objects will be dropped, and all signal handlers removed.
-         *
-         * (If you 
-         */
-        signals[DESTROY] =
-            g_signal_new ("destroy",
-                          HIPPO_TYPE_CANVAS_ITEM,
-                          G_SIGNAL_RUN_CLEANUP | G_SIGNAL_NO_RECURSE | G_SIGNAL_NO_HOOKS,
-                          0, /* No handler; in C implement cleanups in ->dispose() */
-                          NULL, NULL,
-                          g_cclosure_marshal_VOID__VOID,
-                          G_TYPE_NONE, 0);
-            
-        /**
-         * HippoCanvasItem::paint
-         *
-         * All drawing of a canvas item happens in the handlers for
-         * this signal. The rectangle is the bounding box of the
-         * damage region. Most concrete items derive from #HippoCanvasBox,
-         * whose default paint handler invokes a series of more fine-grained
-         * paint handlers to paint the background, content, etc.; usually you
-         * should override one of those fine-grained handlers rather than this
-         * all-encompassing paint.
-         */
-        signals[PAINT] =
-            g_signal_new ("paint",
-                          HIPPO_TYPE_CANVAS_ITEM,
-                          G_SIGNAL_RUN_LAST,
-                          G_STRUCT_OFFSET(HippoCanvasItemIface, paint),
-                          NULL, NULL,
-                          hippo_canvas_marshal_VOID__POINTER_BOXED,
-                          G_TYPE_NONE, 2, G_TYPE_POINTER, HIPPO_TYPE_RECTANGLE);
-        /**
-         * HippoCanvasItem::request-changed
-         *
-         * Signal emitted when the natural or minimum size of the canvas item
-         * may have changed. The parent canvas or parent canvas item will normally
-         * need to recompute its layout in response.
-         */
-        signals[REQUEST_CHANGED] =
-            g_signal_new ("request-changed",
-                          HIPPO_TYPE_CANVAS_ITEM,
-                          G_SIGNAL_RUN_LAST,
-                          G_STRUCT_OFFSET(HippoCanvasItemIface, request_changed),
-                          NULL, NULL,
-                          g_cclosure_marshal_VOID__VOID,
-                          G_TYPE_NONE, 0);
-        /**
-         * HippoCanvasItem::paint-needed
-         *
-         * Signal emitted when a canvas item needs to be repainted. The
-         * rectangle is the bounding box of the areas that need repainting.
-         */
-        signals[PAINT_NEEDED] =
-            g_signal_new ("paint-needed",
-                          HIPPO_TYPE_CANVAS_ITEM,
-                          G_SIGNAL_RUN_LAST,
-                          G_STRUCT_OFFSET(HippoCanvasItemIface, paint_needed),
-                          NULL, NULL,
-                          g_cclosure_marshal_VOID__BOXED,
-                          G_TYPE_NONE, 1, HIPPO_TYPE_RECTANGLE);
-        /**
-         * HippoCanvasItem::button-press-event
-         *
-         * Signal emitted when a mouse button is pressed down on the canvas item.
-         */
-        signals[BUTTON_PRESS_EVENT] =
-            g_signal_new ("button-press-event",
-                          HIPPO_TYPE_CANVAS_ITEM,
-                          G_SIGNAL_RUN_LAST,
-                          G_STRUCT_OFFSET(HippoCanvasItemIface, button_press_event),
-                          g_signal_accumulator_true_handled, NULL,
-                          hippo_canvas_marshal_BOOLEAN__BOXED,
+    /**
+     * HippoCanvasItem::destroy
+     *
+     * This canvas item has explicitely been destroyed. All references it holds
+     * to other objects will be dropped, and all signal handlers removed.
+     */
+    signals[DESTROY] =
+        g_signal_new ("destroy",
+                      HIPPO_TYPE_CANVAS_ITEM,
+                      G_SIGNAL_RUN_CLEANUP | G_SIGNAL_NO_RECURSE | G_SIGNAL_NO_HOOKS,
+                      0, /* No handler; in C implement cleanups in ->dispose() */
+                      NULL, NULL,
+                      g_cclosure_marshal_VOID__VOID,
+                      G_TYPE_NONE, 0);
+    
+    /**
+     * HippoCanvasItem::paint
+     *
+     * All drawing of a canvas item happens in the handlers for
+     * this signal. The rectangle is the bounding box of the
+     * damage region. Most concrete items derive from #HippoCanvasBox,
+     * whose default paint handler invokes a series of more fine-grained
+     * paint handlers to paint the background, content, etc.; usually you
+     * should override one of those fine-grained handlers rather than this
+     * all-encompassing paint.
+     */
+    signals[PAINT] =
+        g_signal_new ("paint",
+                      HIPPO_TYPE_CANVAS_ITEM,
+                      G_SIGNAL_RUN_LAST,
+                      G_STRUCT_OFFSET(HippoCanvasItemIface, paint),
+                      NULL, NULL,
+                      hippo_canvas_marshal_VOID__POINTER_BOXED,
+                      G_TYPE_NONE, 2, G_TYPE_POINTER, HIPPO_TYPE_RECTANGLE);
+    /**
+     * HippoCanvasItem::request-changed
+     *
+     * Signal emitted when the natural or minimum size of the canvas item
+     * may have changed. The parent canvas or parent canvas item will normally
+     * need to recompute its layout in response.
+     */
+    signals[REQUEST_CHANGED] =
+        g_signal_new ("request-changed",
+                      HIPPO_TYPE_CANVAS_ITEM,
+                      G_SIGNAL_RUN_LAST,
+                      G_STRUCT_OFFSET(HippoCanvasItemIface, request_changed),
+                      NULL, NULL,
+                      g_cclosure_marshal_VOID__VOID,
+                      G_TYPE_NONE, 0);
+    /**
+     * HippoCanvasItem::paint-needed
+     *
+     * Signal emitted when a canvas item needs to be repainted. The
+     * rectangle is the bounding box of the areas that need repainting.
+     */
+    signals[PAINT_NEEDED] =
+        g_signal_new ("paint-needed",
+                      HIPPO_TYPE_CANVAS_ITEM,
+                      G_SIGNAL_RUN_LAST,
+                      G_STRUCT_OFFSET(HippoCanvasItemIface, paint_needed),
+                      NULL, NULL,
+                      g_cclosure_marshal_VOID__BOXED,
+                      G_TYPE_NONE, 1, HIPPO_TYPE_RECTANGLE);
+    /**
+     * HippoCanvasItem::button-press-event
+     *
+     * Signal emitted when a mouse button is pressed down on the canvas item.
+     */
+    signals[BUTTON_PRESS_EVENT] =
+        g_signal_new ("button-press-event",
+                      HIPPO_TYPE_CANVAS_ITEM,
+                      G_SIGNAL_RUN_LAST,
+                      G_STRUCT_OFFSET(HippoCanvasItemIface, button_press_event),
+                      g_signal_accumulator_true_handled, NULL,
+                      hippo_canvas_marshal_BOOLEAN__BOXED,
+                      G_TYPE_BOOLEAN, 1, HIPPO_TYPE_EVENT);
+    /**
+     * HippoCanvasItem::button-release-event
+     *
+     * Signal emitted when a mouse button is released on the canvas item.
+     */        
+    signals[BUTTON_RELEASE_EVENT] =
+        g_signal_new ("button-release-event",
+                      HIPPO_TYPE_CANVAS_ITEM,
+                      G_SIGNAL_RUN_LAST,
+                      G_STRUCT_OFFSET(HippoCanvasItemIface, button_release_event),
+                      g_signal_accumulator_true_handled, NULL,
+                      hippo_canvas_marshal_BOOLEAN__BOXED,
+                      G_TYPE_BOOLEAN, 1, HIPPO_TYPE_EVENT);
+    /**
+     * HippoCanvasItem::motion-notify-event
+     *
+     * Signal emitted when the mouse pointer enters, leaves, or moves within
+     * a canvas item. Note that unlike #GtkWidget, there are not separate
+     * events for enter and leave.
+     */                
+    signals[MOTION_NOTIFY_EVENT] =
+        g_signal_new ("motion-notify-event",
+                      HIPPO_TYPE_CANVAS_ITEM,
+                      G_SIGNAL_RUN_LAST,
+                      G_STRUCT_OFFSET(HippoCanvasItemIface, motion_notify_event),
+                      g_signal_accumulator_true_handled, NULL,
+                      hippo_canvas_marshal_BOOLEAN__BOXED,
                       G_TYPE_BOOLEAN, 1, HIPPO_TYPE_EVENT);
-        /**
-         * HippoCanvasItem::button-release-event
-         *
-         * Signal emitted when a mouse button is released on the canvas item.
-         */        
-        signals[BUTTON_RELEASE_EVENT] =
-            g_signal_new ("button-release-event",
-                          HIPPO_TYPE_CANVAS_ITEM,
-                          G_SIGNAL_RUN_LAST,
-                          G_STRUCT_OFFSET(HippoCanvasItemIface, button_release_event),
-                          g_signal_accumulator_true_handled, NULL,
-                          hippo_canvas_marshal_BOOLEAN__BOXED,
-                          G_TYPE_BOOLEAN, 1, HIPPO_TYPE_EVENT);
-        /**
-         * HippoCanvasItem::motion-notify-event
-         *
-         * Signal emitted when the mouse pointer enters, leaves, or moves within
-         * a canvas item. Note that unlike #GtkWidget, there are not separate
-         * events for enter and leave.
-         */                
-        signals[MOTION_NOTIFY_EVENT] =
-            g_signal_new ("motion-notify-event",
-                          HIPPO_TYPE_CANVAS_ITEM,
-                          G_SIGNAL_RUN_LAST,
-                          G_STRUCT_OFFSET(HippoCanvasItemIface, motion_notify_event),
-                          g_signal_accumulator_true_handled, NULL,
-                          hippo_canvas_marshal_BOOLEAN__BOXED,
-                          G_TYPE_BOOLEAN, 1, HIPPO_TYPE_EVENT);
-        /**
-         * HippoCanvasItem::key-press-event
-         *
-         * Signal emitted when a key is pressed while the canvas item is focused.
-         */
-        signals[KEY_PRESS_EVENT] =
-            g_signal_new ("key-press-event",
-                          HIPPO_TYPE_CANVAS_ITEM,
-                          G_SIGNAL_RUN_LAST,
-                          G_STRUCT_OFFSET(HippoCanvasItemIface, key_press_event),
-                          g_signal_accumulator_true_handled, NULL,
-                          hippo_canvas_marshal_BOOLEAN__BOXED,
-                          G_TYPE_BOOLEAN, 1, HIPPO_TYPE_EVENT);
-        /**
-         * HippoCanvasItem::activated
-         *
-         * Signal emitted when the canvas item is "activated" (e.g. if a button is clicked or
-         * an url is clicked).
-         */        
-        signals[ACTIVATED] =
-            g_signal_new ("activated",
-                          HIPPO_TYPE_CANVAS_ITEM,
-                          G_SIGNAL_RUN_LAST,
-                          G_STRUCT_OFFSET(HippoCanvasItemIface, activated),
-                          NULL, NULL,
-                          g_cclosure_marshal_VOID__VOID,
-                          G_TYPE_NONE, 0);
-        /**
-         * HippoCanvasItem::tooltip-changed
-         *
-         * Signal emitted when the canvas item's tooltip changes. The code displaying the
-         * tooltip may need this signal in order to update in response to changes.
-         */                
-        signals[TOOLTIP_CHANGED] =
-            g_signal_new ("tooltip-changed",
-                          HIPPO_TYPE_CANVAS_ITEM,
-                          G_SIGNAL_RUN_LAST,
-                          G_STRUCT_OFFSET(HippoCanvasItemIface, tooltip_changed),
-                          NULL, NULL,
-                          g_cclosure_marshal_VOID__VOID,
-                          G_TYPE_NONE, 0);        
 
+    /**
+     * HippoCanvasItem::scroll-event
+     *
+     * Signal emitted when the mouse wheel or other mechanism requests scrolling.
+     */                
+    signals[SCROLL_EVENT] =
+        g_signal_new ("scroll-event",
+                      HIPPO_TYPE_CANVAS_ITEM,
+                      G_SIGNAL_RUN_LAST,
+                      G_STRUCT_OFFSET(HippoCanvasItemIface, scroll_event),
+                      g_signal_accumulator_true_handled, NULL,
+                      hippo_canvas_marshal_BOOLEAN__BOXED,
+                      G_TYPE_BOOLEAN, 1, HIPPO_TYPE_EVENT);
+        
+    /**
+     * HippoCanvasItem::key-press-event
+     *
+     * Signal emitted when a key is pressed while the canvas item is focused.
+     */
+    signals[KEY_PRESS_EVENT] =
+        g_signal_new ("key-press-event",
+                      HIPPO_TYPE_CANVAS_ITEM,
+                      G_SIGNAL_RUN_LAST,
+                      G_STRUCT_OFFSET(HippoCanvasItemIface, key_press_event),
+                      g_signal_accumulator_true_handled, NULL,
+                      hippo_canvas_marshal_BOOLEAN__BOXED,
+                      G_TYPE_BOOLEAN, 1, HIPPO_TYPE_EVENT);
+    /**
+     * HippoCanvasItem::activated
+     *
+     * Signal emitted when the canvas item is "activated" (e.g. if a button is clicked or
+     * an url is clicked).
+     */        
+    signals[ACTIVATED] =
+        g_signal_new ("activated",
+                      HIPPO_TYPE_CANVAS_ITEM,
+                      G_SIGNAL_RUN_LAST,
+                      G_STRUCT_OFFSET(HippoCanvasItemIface, activated),
+                      NULL, NULL,
+                      g_cclosure_marshal_VOID__VOID,
+                      G_TYPE_NONE, 0);
+    /**
+     * HippoCanvasItem::tooltip-changed
+     *
+     * Signal emitted when the canvas item's tooltip changes. The code displaying the
+     * tooltip may need this signal in order to update in response to changes.
+     */                
+    signals[TOOLTIP_CHANGED] =
+        g_signal_new ("tooltip-changed",
+                      HIPPO_TYPE_CANVAS_ITEM,
+                      G_SIGNAL_RUN_LAST,
+                      G_STRUCT_OFFSET(HippoCanvasItemIface, tooltip_changed),
+                      NULL, NULL,
+                      g_cclosure_marshal_VOID__VOID,
+                      G_TYPE_NONE, 0);
+
+    /**** properties *****/
+        
+    /**
+     * HippoCanvasItem:id
+     * 
+     * The ID of the element; this is used when matching a CSS rule selector against
+     * the element tree.
+     */
+    g_object_interface_install_property(g_iface,
+                                        g_param_spec_string("id",
+                                                            _("ID"),
+                                                            _("ID of the item used for CSS rule matching"),
+                                                            NULL,
+                                                            G_PARAM_READABLE | G_PARAM_WRITABLE));
+        
+    /**
+     * HippoCanvasItem:classes
+     * 
+     * Class names for the element; this is used when matching a CSS rule selector against
+     * the element tree. Multiple classes can be specified, specified by spaces. (This property
+     * is called 'classes' and not 'class' to avoid problems with language bindings.)
+     */
+    g_object_interface_install_property(g_iface,
+                                        g_param_spec_string("classes",
+                                                            _("Classes"),
+                                                            _("Class names for the item used for CSS rule matching"),
+                                                            NULL,
+                                                            G_PARAM_READABLE | G_PARAM_WRITABLE));
         
-        initialized = TRUE;
-    }
 }
 
 void
@@ -328,9 +362,22 @@ hippo_canvas_item_get_tooltip(HippoCanvasItem *canvas_item,
                               int              y,
                               HippoRectangle  *for_area)
 {
+    char *tip;
+    
     g_return_val_if_fail(HIPPO_IS_CANVAS_ITEM(canvas_item), NULL);
 
-    return HIPPO_CANVAS_ITEM_GET_IFACE(canvas_item)->get_tooltip(canvas_item, x, y, for_area);
+    tip = HIPPO_CANVAS_ITEM_GET_IFACE(canvas_item)->get_tooltip(canvas_item, x, y, for_area);
+
+    /* Supporting "" as a synonym for NULL is a convenience for pygtk, which
+     * doesn't have support for NULL-returns from interface methods
+     * at the moment.
+     */
+    if (tip != NULL && strcmp(tip, "") == 0) {
+        g_free(tip);
+        tip = NULL;
+    }
+
+    return tip;
 }
 
 HippoCanvasPointer
@@ -466,6 +513,26 @@ hippo_canvas_item_emit_motion_notify_event (HippoCanvasItem  *canvas_item,
     return result;
 }
 
+gboolean
+hippo_canvas_item_emit_scroll_event (HippoCanvasItem     *canvas_item,
+                                     int                  x,
+                                     int                  y,
+                                     HippoScrollDirection direction)
+{
+    HippoEvent event;
+    gboolean result;
+    
+    g_return_val_if_fail(HIPPO_IS_CANVAS_ITEM(canvas_item), FALSE);
+
+    event.type = HIPPO_EVENT_SCROLL;
+    event.x = x;
+    event.y = y;    
+    event.u.scroll.direction = direction;
+    
+    result = hippo_canvas_item_process_event(canvas_item, &event, 0, 0);
+
+    return result;
+}
 
 gboolean
 hippo_canvas_item_emit_key_press_event (HippoCanvasItem  *canvas_item,
@@ -526,6 +593,10 @@ hippo_canvas_item_emit_paint_needed(HippoCanvasItem *canvas_item,
             damage_box.height = h;
     }
 
+    /* Skip chaining signal emission up the tree if nothing to invalidate */
+    if (damage_box.width <= 0 || damage_box.height <= 0)
+        return;
+    
     g_signal_emit(canvas_item, signals[PAINT_NEEDED], 0,
                   &damage_box);
 }
@@ -582,6 +653,9 @@ hippo_canvas_item_process_event(HippoCanvasItem *canvas_item,
     case HIPPO_EVENT_KEY_PRESS:
         g_signal_emit(canvas_item, signals[KEY_PRESS_EVENT], 0, &translated, &handled);
         break;
+    case HIPPO_EVENT_SCROLL:
+        g_signal_emit(canvas_item, signals[SCROLL_EVENT], 0, &translated, &handled);
+        break;
         /* don't add a default, you'll break the compiler warnings */
     }
 
