@@ -1,6 +1,7 @@
 /* -*- mode: C; c-basic-offset: 4; indent-tabs-mode: nil; -*- */
 #include "hippo-canvas-type-builtins.h"
 #include "hippo-canvas-internal.h"
+#include "hippo-canvas-style.h"
 #include "hippo-canvas-text.h"
 #include "hippo-canvas-box.h"
 #include <pango/pangocairo.h>
@@ -60,8 +61,6 @@ enum {
     PROP_SIZE_MODE
 };
 
-#define DEFAULT_FOREGROUND 0x000000ff
-
 G_DEFINE_TYPE_WITH_CODE(HippoCanvasText, hippo_canvas_text, HIPPO_TYPE_CANVAS_BOX,
                         G_IMPLEMENT_INTERFACE(HIPPO_TYPE_CANVAS_ITEM, hippo_canvas_text_iface_init));
 
@@ -234,6 +233,7 @@ hippo_canvas_text_set_property(GObject         *object,
                 g_free(text->text);
                 text->text = g_strdup(new_text);
                 hippo_canvas_item_emit_request_changed(HIPPO_CANVAS_ITEM(text));
+                hippo_canvas_item_emit_paint_needed(HIPPO_CANVAS_ITEM(text), 0, 0, -1, -1);
             }
         }
         break;
@@ -246,6 +246,7 @@ hippo_canvas_text_set_property(GObject         *object,
                 pango_attr_list_unref(text->attributes);            
             text->attributes = attrs;
             hippo_canvas_item_emit_request_changed(HIPPO_CANVAS_ITEM(text));
+            hippo_canvas_item_emit_paint_needed(HIPPO_CANVAS_ITEM(text), 0, 0, -1, -1);
         }
         break;
     case PROP_MARKUP:
@@ -272,10 +273,12 @@ hippo_canvas_text_set_property(GObject         *object,
     case PROP_FONT_SCALE:
         text->font_scale = g_value_get_double(value);
         hippo_canvas_item_emit_request_changed(HIPPO_CANVAS_ITEM(text));
+        hippo_canvas_item_emit_paint_needed(HIPPO_CANVAS_ITEM(text), 0, 0, -1, -1);
         break;
     case PROP_SIZE_MODE:
         text->size_mode = g_value_get_enum(value);
         hippo_canvas_item_emit_request_changed(HIPPO_CANVAS_ITEM(text));
+        hippo_canvas_item_emit_paint_needed(HIPPO_CANVAS_ITEM(text), 0, 0, -1, -1);
         break;
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
@@ -327,8 +330,10 @@ hippo_canvas_text_set_context(HippoCanvasItem    *item,
      * so we have to queue a size change when the context
      * is set.
      */
-    if (changed)
+    if (changed) {
         hippo_canvas_item_emit_request_changed(HIPPO_CANVAS_ITEM(item));
+        hippo_canvas_item_emit_paint_needed(HIPPO_CANVAS_ITEM(item), 0, 0, -1, -1);
+    }
 }
 
 
@@ -373,51 +378,33 @@ static PangoLayout*
 create_layout(HippoCanvasText *text,
               int              allocation_width)
 {
-    HippoCanvasContext *context;
+    HippoCanvasBox *box = HIPPO_CANVAS_BOX(text);
     PangoLayout *layout;
+    HippoCanvasStyle *style = hippo_canvas_context_get_style(HIPPO_CANVAS_CONTEXT(text));
 
-    /* Note that our context is *ourselves* not box->context i.e. we want
-     * our own style, etc. to affect what we render. Our context methods
-     * will chain up as needed.
-     */
-    context = HIPPO_CANVAS_CONTEXT(text);
-
-    g_return_val_if_fail(context != NULL, NULL);
+    g_return_val_if_fail(box->context != NULL, NULL);
     
-    layout = hippo_canvas_context_create_layout(context);
+    layout = hippo_canvas_context_create_layout(box->context);
 
-    {
-        const PangoFontDescription *old;
-        PangoFontDescription *composite;
-        
-        composite = pango_font_description_new();
-        
-        old = pango_layout_get_font_description(layout);
-        /* if no font desc is set on the layout, the layout uses the one
-         * from the context, so emulate that here.
-         */
-        if (old == NULL)
-            old = pango_context_get_font_description(pango_layout_get_context(layout));
-        
-        if (old != NULL)
-            pango_font_description_merge(composite, old, TRUE);
-
-        hippo_canvas_context_affect_font_desc(context,
-                                              composite);
-        
-        pango_layout_set_font_description(layout, composite);
-    
-        pango_font_description_free(composite);
+    if (box->font_desc) {
+        PangoFontDescription *merged = pango_font_description_copy(hippo_canvas_style_get_font(style));
+        pango_font_description_merge(merged, box->font_desc, TRUE);
+        pango_layout_set_font_description(layout, merged);
+        pango_font_description_free(merged);
+    } else {
+        pango_layout_set_font_description(layout, hippo_canvas_style_get_font(style));
     }
     
     {
         PangoAttrList *attrs;
+        HippoTextDecoration decoration = hippo_canvas_style_get_text_decoration(style);
         
         if (text->attributes)
             attrs = pango_attr_list_copy(text->attributes);
         else
             attrs = pango_attr_list_new();
 
+        
         if (ABS(1.0 - text->font_scale) > .000001) {
             PangoAttribute *attr = pango_attr_scale_new(text->font_scale);
             attr->start_index = 0;
@@ -425,6 +412,20 @@ create_layout(HippoCanvasText *text,
             pango_attr_list_insert(attrs, attr);
         }
 
+        if ((decoration & HIPPO_TEXT_DECORATION_UNDERLINE) != 0) {
+            PangoAttribute *attr = pango_attr_underline_new(TRUE);
+            attr->start_index = 0;
+            attr->end_index = G_MAXUINT;
+            pango_attr_list_insert(attrs, attr);
+        }
+
+        if ((decoration & HIPPO_TEXT_DECORATION_LINE_THROUGH) != 0) {
+            PangoAttribute *attr = pango_attr_strikethrough_new(TRUE);
+            attr->start_index = 0;
+            attr->end_index = G_MAXUINT;
+            pango_attr_list_insert(attrs, attr);
+        }
+
         pango_layout_set_attributes(layout, attrs);
         pango_attr_list_unref(attrs);
     }
@@ -555,14 +556,14 @@ hippo_canvas_text_paint_below_children(HippoCanvasBox  *box,
 {
     HippoCanvasText *text = HIPPO_CANVAS_TEXT(box);
     guint32 color_rgba;
-    HippoCanvasContext *context;
 
-    /* note, we want to use _ourselves_ as context so we get our own style */
-    context = HIPPO_CANVAS_CONTEXT(box);
-
-    color_rgba = HIPPO_CANVAS_BOX_GET_CLASS(box)->default_color;
-    hippo_canvas_context_affect_color(context, &color_rgba);
-    
+    if (box->color_set) {
+        color_rgba = box->color_rgba;
+    } else {
+        HippoCanvasStyle *style = hippo_canvas_context_get_style(HIPPO_CANVAS_CONTEXT(text));
+        color_rgba = hippo_canvas_style_get_foreground_color(style);
+    }
+        
     /* It would seem more natural to compute whether we are ellipsized or
      * not when we are allocated to some width, but we don't have a layout
      * at that point. We could do it in get_content_height_request(), but
