@@ -11,6 +11,9 @@
 #include <gtk/gtkwindow.h>
 #include <gtk/gtklabel.h>
 
+#include <gdk/gdkx.h>
+#include <cairo-xlib.h>
+
 /* Gap between the area we are tipping and the tooltip */
 #define TOOLTIP_PADDING 4
 /* Maximum width of a tippable area before we position the tooltip by the pointer */
@@ -43,6 +46,10 @@ static void hippo_canvas_helper_get_property (GObject      *object,
                                               GParamSpec   *pspec);
 
 static PangoLayout*     hippo_canvas_helper_create_layout          (HippoCanvasContext *context);
+static cairo_surface_t* hippo_canvas_helper_create_surface         (HippoCanvasContext *context,
+                                                                    cairo_content_t     content,
+                                                                    int                 width,
+                                                                    int                 height);
 static cairo_surface_t* hippo_canvas_helper_load_image             (HippoCanvasContext *context,
                                                                     const char         *image_name);
 static guint32          hippo_canvas_helper_get_color              (HippoCanvasContext *context,
@@ -64,8 +71,14 @@ HippoCanvasStyle *    hippo_canvas_helper_get_style      (HippoCanvasContext *co
 double                hippo_canvas_helper_get_resolution (HippoCanvasContext *context);
 PangoFontDescription *hippo_canvas_helper_get_font       (HippoCanvasContext *context);
 
+HippoAnimationManager *hippo_canvas_helper_get_animation_manager   (HippoCanvasContext *context);
+
 static void             hippo_canvas_helper_fixup_resize_state     (HippoCanvasHelper  *canvas);
 
+static void       on_animation_manager_after_frame (HippoAnimationManager *manager,
+                                                    guint                  frame_serial,
+                                                    HippoCanvasHelper     *helper);
+
 static void       tooltip_window_update   (GtkWidget      *tip,
                                            GdkScreen      *screen,
                                            int             mouse_x,
@@ -100,8 +113,13 @@ struct _HippoCanvasHelper {
     
     GSList *widget_items;
 
+    HippoAnimationManager *animation_manager;
+    guint frame_serial;
+
     unsigned int root_hovering : 1;
     unsigned int fixing_up_resize_state : 1;
+    unsigned int need_background_paint : 1;
+    unsigned int frame_pending : 1;
 };
 
 struct _HippoCanvasHelperClass {
@@ -130,6 +148,10 @@ hippo_canvas_helper_init(HippoCanvasHelper *helper)
     helper->pointer = HIPPO_CANVAS_POINTER_UNSET;
     helper->last_window_x = -1;
     helper->last_window_y = -1;
+
+    helper->animation_manager = hippo_animation_manager_new();
+    g_signal_connect(helper->animation_manager, "after-frame",
+                     G_CALLBACK(on_animation_manager_after_frame), helper);
 }
 
 static void
@@ -148,6 +170,7 @@ static void
 hippo_canvas_helper_iface_init (HippoCanvasContextIface *klass)
 {
     klass->create_layout = hippo_canvas_helper_create_layout;
+    klass->create_surface = hippo_canvas_helper_create_surface;
     klass->load_image = hippo_canvas_helper_load_image;
     klass->get_color = hippo_canvas_helper_get_color;
     klass->register_widget_item = hippo_canvas_helper_register_widget_item;
@@ -157,6 +180,7 @@ hippo_canvas_helper_iface_init (HippoCanvasContextIface *klass)
     klass->get_style = hippo_canvas_helper_get_style;
     klass->get_resolution = hippo_canvas_helper_get_resolution;
     klass->get_font = hippo_canvas_helper_get_font;
+    klass->get_animation_manager = hippo_canvas_helper_get_animation_manager;
 }
 
 static void
@@ -191,14 +215,23 @@ hippo_canvas_helper_dispose(GObject *object)
         gtk_object_destroy(GTK_OBJECT(helper->tooltip_window));
         helper->tooltip_window = NULL;
     }
-    
+
+    if (helper->animation_manager) {
+        g_signal_handlers_disconnect_by_func(helper->animation_manager,
+                                             (gpointer)on_animation_manager_after_frame,
+                                             helper);
+        
+        g_object_unref(helper->animation_manager);
+        helper->animation_manager = NULL;
+    }
+
     G_OBJECT_CLASS(hippo_canvas_helper_parent_class)->dispose(object);
 }
 
 static void
 hippo_canvas_helper_finalize(GObject *object)
 {
-    /* HippoCanvasHelper *helper = HIPPO_CANVAS(object); */
+    /* HippoCanvasHelper *helper = HIPPO_CANVAS_HELPER(object); */
 
     G_OBJECT_CLASS(hippo_canvas_helper_parent_class)->finalize(object);
 }
@@ -373,29 +406,102 @@ update_tooltip(HippoCanvasHelper *helper,
     }
 }
 
+static gboolean
+allocate_pending_on_widget(GtkWidget *widget)
+{
+    return GTK_WIDGET_ALLOC_NEEDED(widget);
+}
+
+static gboolean
+expose_pending_on_window(GdkWindow *window)
+{
+    /* This will not work properly for the OS X port of GTK+ */
+
+    return ((GdkWindowObject *)window)->update_area != NULL;
+}
+
+static void
+on_animation_manager_after_frame (HippoAnimationManager *manager,
+                                  guint                  frame_serial,
+                                  HippoCanvasHelper     *helper)
+{
+    /* A frame has finished; we now need to wait until all resizing
+     * and redrawing and then call
+     * hippo_animation_manager_frame_complete().
+     */
+    
+    if (!allocate_pending_on_widget(helper->widget) &&
+        !expose_pending_on_window(helper->widget->window))
+    {
+        /* Nothing to do, we can call frame_complete() immediately */
+        hippo_animation_manager_frame_complete(manager, frame_serial);
+    }
+    else
+    {
+        helper->frame_serial = frame_serial;
+        helper->frame_pending = TRUE;
+    }
+}
+
 gboolean
 hippo_canvas_helper_expose_event(HippoCanvasHelper *helper,
                                  GdkEventExpose    *event)
 {
-    
     cairo_t *cr;
-    int window_x, window_y;
-    HippoRectangle damage_box;
-
-    if (helper->root == NULL)
-        return FALSE;
 
     cr = gdk_cairo_create(event->window);
-    get_root_item_window_coords(helper, &window_x, &window_y);
 
-    damage_box.x = event->area.x;
-    damage_box.y = event->area.y;
-    damage_box.width = event->area.width;
-    damage_box.height = event->area.height;
-    hippo_canvas_item_process_paint(helper->root, cr, &damage_box,
-                                    window_x, window_y);
+    if (helper->need_background_paint) {
+        HippoCanvasStyle *style = hippo_canvas_context_get_style(HIPPO_CANVAS_CONTEXT(helper));
+        guint32 color = hippo_canvas_style_get_background_color(style);
+
+        cairo_save(cr);
+        cairo_set_source_rgba(cr,
+                              ((color & 0xFF000000) >> 24) / 255.,
+                              ((color & 0x00FF0000) >> 16) / 255.,
+                              ((color & 0x0000FF00) >> 8) / 255.,
+                              ((color & 0x000000FF)) / 255.);
+        cairo_set_operator(cr, CAIRO_OPERATOR_SOURCE);
+        cairo_paint(cr);
+        cairo_restore(cr);
+    }
+
+    if (helper->root != NULL) {
+        int window_x, window_y;
+        HippoRectangle damage_box;
+    
+        get_root_item_window_coords(helper, &window_x, &window_y);
+
+        damage_box.x = event->area.x;
+        damage_box.y = event->area.y;
+        damage_box.width = event->area.width;
+        damage_box.height = event->area.height;
+        hippo_canvas_item_process_paint(helper->root, cr, &damage_box,
+                                        window_x, window_y);
+    }
+    
     cairo_destroy(cr);
 
+    if (helper->frame_pending) {
+        /* Resize occurs before redraw, so when we are done redrawing, we know we are
+         * done with the frame
+         */
+
+        /* In theory, we should actually wait for the frame to paint on the X server;
+         * we could do this by changing a property and watching for the PropertyNotify
+         * event or by using the XSync extension. Even better would be to communicate
+         * with the compositing manager and wait until a frame containing the new
+         * contents of our window has been drawn to the front buffer.
+         *
+         * In practice, the difference between that and just going ahead and starting
+         * the next frame is going to be pretty small unless the balance
+         * of server vs. client side efficiency changes a lot for cairo... the client
+         * side of cairo currently doesn't keep up with the server.
+         */
+        helper->frame_pending = FALSE;
+        hippo_animation_manager_frame_complete(helper->animation_manager, helper->frame_serial);
+    }
+    
     return FALSE;
 }
 
@@ -432,15 +538,12 @@ hippo_canvas_helper_size_allocate(HippoCanvasHelper *helper,
         int border_width = GTK_CONTAINER(helper->widget)->border_width;
         int child_width = allocation->width - border_width * 2;
         int child_height = allocation->height - border_width * 2;
-        int old_width, old_height;
         gboolean border_changed;
         
-        hippo_canvas_item_get_allocation(helper->root, &old_width, &old_height);
-
         border_changed = border_width = helper->last_allocated_border;
         helper->last_allocated_border = border_width;
 
-        if (child_width != old_width || child_height != old_height || border_changed)
+        if (border_changed)
             gtk_widget_queue_draw(helper->widget);
         
         hippo_canvas_item_allocate(helper->root,
@@ -450,6 +553,15 @@ hippo_canvas_helper_size_allocate(HippoCanvasHelper *helper,
         /* Tooltip might be in the wrong place now */
         update_tooltip(helper, FALSE);
     }
+
+    if (helper->frame_pending) {
+        if (!helper->widget->window || !expose_pending_on_window(helper->widget->window)) {
+            /* We resized, and there was nothing to draw, we can can call frame_complete() */
+            
+            helper->frame_pending = FALSE;
+            hippo_animation_manager_frame_complete(helper->animation_manager, helper->frame_serial);
+        }
+    }
 }
 
 gboolean
@@ -698,6 +810,14 @@ hippo_canvas_helper_hierarchy_changed (HippoCanvasHelper *helper,
     cancel_tooltip(helper);
 }
 
+static int
+premultiply(int component,
+            int alpha)
+{
+    int temp = component * alpha + 0x80;
+    return ((temp >> 8) + temp) >> 8;
+}
+
 void
 hippo_canvas_helper_set_window_background (HippoCanvasHelper *helper,
                                            GdkWindow         *window)
@@ -705,18 +825,51 @@ hippo_canvas_helper_set_window_background (HippoCanvasHelper *helper,
     HippoCanvasStyle *style = hippo_canvas_context_get_style(HIPPO_CANVAS_CONTEXT(helper));
     guint32 color;
     
+    helper->need_background_paint = FALSE;
+
+    /* We don't use hippo_canvas_style_get_background_color() here because we treat
+     * background-color: transparent on the canvas as being different from
+     * not specifying it ... which should give us the theme color. Perhaps
+     * a cleaner thing to do would be to support the CSS2 system colors
+     * and have a default stylesheet with canvas { background-color: Window; }
+     */
     if (hippo_canvas_style_get_color(style, "background-color", FALSE, &color)) {
         GdkColormap *colormap = gdk_window_get_colormap(window);
+        GdkVisual *visual = gdk_colormap_get_visual(colormap);
         GdkColor color_gdk;
 
-        color_gdk.red =   (color & 0xFF000000) >> 24;
-        color_gdk.red = color_gdk.red * 0x101;
-        color_gdk.green = (color & 0x00FF0000) >> 16;
-        color_gdk.green = color_gdk.green * 0x101;
-        color_gdk.blue =  (color & 0x0000FF00) >> 8;
-        color_gdk.blue = color_gdk.blue * 0x101;
+        /* Special-case the RGBA visual. If visuals other than ARGB32 were in use
+         * (say ABGR32) then we should support them here as well.
+         */
+        if (visual->depth == 32 &&
+            visual->red_mask == 0xff0000 &&
+            visual->green_mask == 0x00ff00 &&
+            visual->blue_mask  == 0x0000ff)
+        {
+            int alpha = color & 0x000000FF;
 
-        gdk_rgb_find_color(colormap, &color_gdk);
+            if (alpha != 0xFF)
+                /* GTK+ (as of 2.12) doesn't handle clearing to a RGBA color correctly
+                 * so we'll need to fix up the double-buffer pixmap ourself.
+                 */
+                helper->need_background_paint = TRUE;
+
+            color_gdk.pixel = ((alpha << 24) |
+                               (premultiply((color & 0xFF000000) >> 24, alpha) << 16) |
+                               (premultiply((color & 0x00FF0000) >> 16, alpha) << 8) |
+                               (premultiply((color & 0x0000FF00) >> 8,  alpha)));
+	}
+        else
+        {
+            color_gdk.red =   (color & 0xFF000000) >> 24;
+            color_gdk.red = color_gdk.red * 0x101;
+            color_gdk.green = (color & 0x00FF0000) >> 16;
+            color_gdk.green = color_gdk.green * 0x101;
+            color_gdk.blue =  (color & 0x0000FF00) >> 8;
+            color_gdk.blue = color_gdk.blue * 0x101;
+            
+            gdk_rgb_find_color(colormap, &color_gdk);
+        }
 
         gdk_window_set_background(window, &color_gdk);
     } else {
@@ -790,6 +943,36 @@ hippo_canvas_helper_create_layout(HippoCanvasContext *context)
     return gtk_widget_create_pango_layout(helper->widget, NULL);
 }
 
+static cairo_surface_t*
+hippo_canvas_helper_create_surface(HippoCanvasContext *context,
+                                   cairo_content_t     content,
+                                   int                 width,
+                                   int                 height)
+{
+    /* This feels like a really round-about way of doing things, but
+     * I can't think of a better way. There's no way to get a cairo
+     * xlib surface without a visual other than create_similar(), and
+     * no visual for CAIRO_FORMAT_ALPHA.
+     */
+    HippoCanvasHelper *helper = HIPPO_CANVAS_HELPER(context);
+    GdkScreen *screen = gtk_widget_get_screen(helper->widget);
+    GdkDisplay *display = gdk_screen_get_display(screen);
+    GdkWindow *root = gdk_screen_get_root_window(screen);
+    GdkVisual *visual = gdk_drawable_get_visual(root);
+    cairo_surface_t *root_surface;
+    cairo_surface_t *surface;
+
+    root_surface = cairo_xlib_surface_create(GDK_DISPLAY_XDISPLAY(display),
+                                             GDK_DRAWABLE_XID(root),
+                                             GDK_VISUAL_XVISUAL(visual),
+                                             gdk_screen_get_width(screen),
+                                             gdk_screen_get_height(screen));
+    surface = cairo_surface_create_similar(root_surface, content, width, height);
+    cairo_surface_destroy(root_surface);
+
+    return surface;
+}
+
 static HippoCanvasLoadImageHook hippo_canvas_helper_load_image_hook = NULL;
 
 void
@@ -1061,6 +1244,14 @@ hippo_canvas_helper_get_font (HippoCanvasContext *context)
     return helper->widget->style->font_desc;
 }
 
+HippoAnimationManager *
+hippo_canvas_helper_get_animation_manager (HippoCanvasContext *context)
+{
+    HippoCanvasHelper *helper = HIPPO_CANVAS_HELPER(context);
+
+    return helper->animation_manager;
+}
+
 static void
 canvas_root_destroy(HippoCanvasItem   *root,
                     HippoCanvasHelper *helper)
@@ -1087,7 +1278,7 @@ canvas_root_paint_needed(HippoCanvasItem      *root,
     int window_x, window_y;
     
     get_root_item_window_coords(helper, &window_x, &window_y);
-    
+
     gtk_widget_queue_draw_area(widget,
                                damage_box->x + window_x,
                                damage_box->y + window_y,
